import pytest
from mock import Mock
from hamcrest import assert_that, instance_of, equal_to, starts_with
from hamcrest.core.string_description import StringDescription
from matchmock import (
    called, not_called, called_once, called_with, called_n_times)


@pytest.fixture
def desc():
    return StringDescription()


def test_called_ok() -> None:
    matcher = called()
    mock = Mock()
    mock()

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_called_member_ok() -> None:
    matcher = called()
    mock = Mock()
    mock.fun()

    ok = matcher.matches(mock.fun)
    assert_that(ok, equal_to(True))


def test_called_ok_twice() -> None:
    matcher = called()
    mock = Mock()
    mock()
    mock()

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_called_ok_explicitly_twice() -> None:
    matcher = called_n_times(2)
    mock = Mock()
    mock()
    mock()

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_called_ok_explicitly_twice_fail(desc) -> None:
    expected = 'was called 3 times'
    matcher = called_n_times(2)
    mock = Mock()
    mock()
    mock()
    mock()

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), starts_with(expected))


def test_called_description(desc) -> None:
    expected = 'Mock called a value greater than <0> times with ANYTHING'
    matcher = called()
    matcher.describe_to(desc)
    assert_that(str(desc), equal_to(expected))


def test_called_fail(desc) -> None:
    expected = 'was called 0 times'
    matcher = called()
    mock = Mock()

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_not_called_ok() -> None:
    matcher = not_called()
    mock = Mock()

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_not_called_description(desc) -> None:
    matcher = not_called()
    expected = 'Mock called <0> times with ANYTHING'
    matcher.describe_to(desc)
    assert_that(str(desc), equal_to(expected))


def test_not_called_fail(desc) -> None:
    matcher = not_called()
    expected = "was called 1 times with ('foo', )"

    mock = Mock()
    mock('foo')

    ok = matcher.matches(mock, desc)

    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_once_ok() -> None:
    matcher = called_once()
    mock = Mock()
    mock('foo')

    ok = matcher.matches(mock)

    assert_that(ok, equal_to(True))


def test_called_once_description(desc) -> None:
    expected = 'Mock called <1> times with ANYTHING'
    matcher = called_once()
    matcher.describe_to(desc)
    assert_that(str(desc), equal_to(expected))


def test_called_once_fail_not_called(desc) -> None:
    expected = 'was called 0 times'
    matcher = called_once()
    mock = Mock()

    ok = matcher.matches(mock, desc)

    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_once_fail_called_twice(desc) -> None:
    expected = "was called 2 times with ('foo', ) and ('bar', )"
    matcher = called_once()
    mock = Mock()
    mock('foo')
    mock('bar')

    ok = matcher.matches(mock, desc)

    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_with_ok() -> None:
    matcher = called_with('foo', instance_of(int))
    mock = Mock()
    mock('foo', 5)

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_called_with_ok_kwargs() -> None:
    matcher = called_with(x='foo', y=instance_of(int))
    mock = Mock()
    mock(x='foo', y=5)

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_called_with_ok_multi() -> None:
    matcher = called_with('foo', instance_of(int))
    mock = Mock()
    mock('baz')
    mock('foo', 5)
    mock('bar', 5)

    ok = matcher.matches(mock)
    assert_that(ok, equal_to(True))


def test_called_with_description(desc) -> None:
    expected = ("Mock called a value greater than <0> times with " +
                "('foo', an instance of int, )")
    matcher = called_with('foo', instance_of(int))

    matcher.describe_to(desc)
    assert_that(str(desc), equal_to(expected))


def test_called_with_failed(desc) -> None:
    expected = "argument 1: was 'bar'"
    matcher = called_with('foo', instance_of(int))
    mock = Mock()
    mock('foo', 'bar')

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_with_failed_multi(desc) -> None:
    expected = "in call 0: argument 1: was 'bar', in call 1: 1 extra arguments"
    matcher = called_with('foo', instance_of(int))
    mock = Mock()
    mock('foo', 'bar')
    mock('foo', 'bar', 'baz')

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_with_kwarg_failed(desc) -> None:
    expected = "value for 'foo' was 'baz'"
    matcher = called_with(foo='bar')
    mock = Mock()
    mock(foo='baz')

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_with_extra_arg(desc) -> None:
    expected = "1 extra arguments"
    matcher = called_with('spam')
    mock = Mock()
    mock('spam', 'bar')

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))


def test_called_with_extra_kwarg(desc) -> None:
    expected = "extra keyword argument(s) 'foo' given"
    matcher = called_with('spam')
    mock = Mock()
    mock('spam', foo='bar')

    ok = matcher.matches(mock, desc)
    assert_that(ok, equal_to(False))
    assert_that(str(desc), equal_to(expected))



'''Hamcrest matchers for mock objects'''

from typing import Any, Collection, Mapping, Optional, Sequence, Tuple

from unittest.mock import Mock, _Call
from hamcrest.core.base_matcher import BaseMatcher
from hamcrest.core.description import Description
from hamcrest.core.matcher import Matcher
from hamcrest.core.helpers.wrap_matcher import wrap_matcher
from hamcrest import (
    equal_to, anything, has_entries, has_item, greater_than)

__all__ = ['called', 'not_called', 'called_once',
           'called_with', 'called_once_with', 'called_n_times']

_Args = Tuple
_Kwargs = Mapping[str, Any]


def describe_call(args: _Args, kwargs: _Kwargs, desc: Description) -> None:
    desc.append_text('(')
    desc.append_list('', ', ', '', args)
    desc.append_text(', ')
    first = True
    for key, value in sorted(kwargs.items()):
        if not first:
            desc.append_text(', ')
        desc.append_text(key)   \
            .append_text('=')  \
            .append_description_of(value)
        first = False
    desc.append_text(')')


class IsCall(BaseMatcher[_Call]):
    '''A matcher that describes a call.

    The positional arguments and keyword arguments are represented with
    individual submatchers.
    '''
    args: Matcher[_Args]
    kwargs: Matcher[_Kwargs]

    def __init__(self, args: Matcher[_Args], kwargs: Matcher[_Kwargs]) -> None:
        self.args = args
        self.kwargs = kwargs

    def _matches(self, item: _Call) -> bool:
        # in python >= 3.8 this can be item.args, item.kwargs
        if len(item) == 3:
            _name, args, kwargs = item
        else:
            args, kwargs = item
        return self.args.matches(args) and self.kwargs.matches(kwargs)

    def describe_mismatch(
        self,
        item: _Call,
        mismatch_description: Description
    ) -> None:
        # in python >= 3.8 this can be item.args, item.kwargs
        if len(item) == 3:
            _name, args, kwargs = item
        else:
            args, kwargs = item
        if self.args.matches(args, mismatch_description):
            self.kwargs.matches(kwargs, mismatch_description)

    def describe_to(self, desc: Description) -> None:
        desc.append_text('(')
        desc.append_description_of(self.args)
        desc.append_text(', ')
        desc.append_description_of(self.kwargs)
        desc.append_text(')')


class IsArgs(BaseMatcher[_Args]):
    matchers: Sequence[Matcher]

    def __init__(self, matchers: Sequence[Matcher]) -> None:
        self.matchers = matchers

    def matches(
        self,
        obj: _Args,
        mismatch_description: Optional[Description] = None
    ) -> bool:
        md = mismatch_description
        if len(obj) < len(self.matchers):
            if md:
                d = len(self.matchers) - len(obj)
                md.append_text('missing %s arguments' % d)
            return False
        if len(obj) > len(self.matchers):
            if md:
                d = len(obj) - len(self.matchers)
                md.append_text('%s extra arguments' % d)
            return False
        for i, (o, m) in enumerate(zip(obj, self.matchers)):
            if not m.matches(o):
                if md:
                    md.append_text('argument ' + str(i) + ': ')
                    m.describe_mismatch(o, md)
                return False
        return True

    def describe_to(self, desc: Description) -> None:
        desc.append_list('', ', ', '', self.matchers)


class IsKwargs(BaseMatcher[_Kwargs]):
    _value_matchers: Collection[Tuple[str, Matcher]]
    _matcher: Matcher[_Kwargs]

    def __init__(self, value_matchers: Mapping[str, Matcher]) -> None:
        self._value_matchers = value_matchers.items()
        self._matcher = has_entries(value_matchers)

    def matches(
        self,
        obj: _Kwargs,
        mismatch_description: Optional[Description] = None
    ) -> bool:
        md = mismatch_description
        ok = self._matcher.matches(obj, md)
        if not ok:
            return False

        expected_keys = set(k for k, _ in self._value_matchers)
        actual_keys = set(obj.keys())
        extra_keys = actual_keys - expected_keys
        if len(extra_keys) > 0:
            if md:
                md.append_text('extra keyword argument(s) ') \
                               .append_list('', ', ', '', extra_keys) \
                               .append_text(' given')
            return False

        return True

    def describe_to(self, desc: Description) -> None:
        first = True
        for key, value in self._value_matchers:
            if not first:
                desc.append_text(', ')
            desc.append_text(key)   \
                .append_text('=')  \
                .append_description_of(value)
            first = False


def match_args(args: Sequence[Any]) -> Matcher[_Args]:
    '''Create a matcher for positional arguments'''

    return IsArgs(tuple(wrap_matcher(m) for m in args))


def match_kwargs(kwargs: Mapping[str, Any]) -> Matcher[_Kwargs]:
    '''Create a matcher for keyword arguments'''

    return IsKwargs({k: wrap_matcher(v) for k, v in kwargs.items()})


class IsCalled(BaseMatcher[Mock]):
    '''Matches a mock and asserts the number of calls and parameters'''

    call: Matcher[_Call]
    count: Matcher[int]
    has_call: Matcher[Sequence[_Call]]

    def __init__(self, call: Matcher[_Call], count: Matcher[int]) -> None:
        self.call = call
        self.count = count
        self.has_call = has_item(self.call)

    def _matches(self, item: Mock) -> bool:
        if not self.count.matches(item.call_count):
            return False

        if len(item.call_args_list) == 0:
            return True

        return self.has_call.matches(item.call_args_list)

    def describe_mismatch(
        self,
        item: Mock,
        mismatch_description: Description
    ) -> None:
        if not self.count.matches(item.call_count):
            mismatch_description.append_text(
                'was called %s times' % item.call_count)
            if item.call_count > 0:
                mismatch_description.append_text(' with ')

                for i, call in enumerate(item.call_args_list):
                    if i != 0:
                        mismatch_description.append_text(' and ')
                    describe_call(call[0], call[1], mismatch_description)
            return

        if item.call_count > 0:
            for i, call in enumerate(item.call_args_list):
                if item.call_count > 1:
                    mismatch_description.append_text('in call %s: ' % i)
                self.call.describe_mismatch(call, mismatch_description)
                if i != item.call_count - 1:
                    mismatch_description.append_text(', ')

    def describe_to(self, desc: Description) -> None:
        desc.append_text('Mock called ')
        self.count.describe_to(desc)
        desc.append_text(' times with ')
        self.call.describe_to(desc)


def called() -> Matcher[Mock]:
    '''Match mock that was called one or more times'''

    return IsCalled(anything(), count=greater_than(0))


def called_n_times(n) -> Matcher[Mock]:
    '''Match mock that was called exactly a given number of times'''

    return IsCalled(anything(), count=equal_to(n))


def not_called() -> Matcher[Mock]:
    '''Match mock that was never called'''

    return IsCalled(anything(), count=equal_to(0))


def called_once() -> Matcher[Mock]:
    '''Match mock that was called once regardless of arguments'''

    return IsCalled(anything(), count=equal_to(1))


def called_with(*args, **kwargs) -> Matcher[Mock]:
    '''Match mock has at least one call with the specified arguments'''

    return IsCalled(IsCall(match_args(args), match_kwargs(kwargs)),
                    count=greater_than(0))


def called_once_with(*args, **kwargs) -> Matcher[Mock]:
    '''Match mock that was called once and with the specified arguments'''

    return IsCalled(IsCall(match_args(args), match_kwargs(kwargs)),
                    count=equal_to(1))



from hamcrest import assert_that, equal_to, has_entries
from hamcrest.core.string_description import StringDescription
from matchmock import IsCall, match_args, match_kwargs
from unittest.mock import call


def test_matching_call() -> None:
    m = IsCall(match_args(('foo',)), match_kwargs({}))
    value = call('foo')
    assert_that(m.matches(value), equal_to(True))


def test_mismatching_call() -> None:
    m = IsCall(match_args(('foo',)), match_kwargs({}))
    value = call('bar')
    assert_that(m.matches(value), equal_to(False))


def test_describe_self() -> None:
    m = IsCall(match_args(('foo',)), match_kwargs({}))
    s = StringDescription()
    m.describe_to(s)
    assert_that(str(s), equal_to("('foo', )"))


def test_describe_self_with_kwargs() -> None:
    m = IsCall(match_args(('foo',)), match_kwargs({'key': 'value'}))
    s = StringDescription()
    m.describe_to(s)
    assert_that(str(s), equal_to("('foo', key='value')"))


def test_describe_mismatch() -> None:
    m = IsCall(match_args(('foo',)), match_kwargs({}))
    value = call('bar')
    s = StringDescription()
    m.describe_mismatch(value, s)
    assert_that(str(s), equal_to("argument 0: was 'bar'"))


def test_describe_mismatch_kwargs() -> None:
    m = IsCall(match_args(('foo',)), match_kwargs({'key': 'value'}))
    value = call('foo', key='VALUE')
    s = StringDescription()
    m.describe_mismatch(value, s)
    assert_that(str(s), equal_to("value for 'key' was 'VALUE'"))


def test_args_mismatch_complex() -> None:
    m = IsCall(match_args([has_entries(name='foo')]), match_kwargs({}))
    value = call({'name': 'baz'})
    s = StringDescription()
    m.matches(value, s)
    assert_that(str(s), equal_to("argument 0: value for 'name' was 'baz'"))
